"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.SshTunnel = void 0;
const util_1 = require("util");
const events_1 = require("events");
const ssh2_1 = require("ssh2");
const compass_logging_1 = require("@mongodb-js/compass-logging");
const server_1 = __importDefault(require("socksv5/lib/server"));
const None_1 = __importDefault(require("socksv5/lib/auth/None"));
const UserPassword_1 = __importDefault(require("socksv5/lib/auth/UserPassword"));
const { log, mongoLogId, debug } = (0, compass_logging_1.createLoggerAndTelemetry)('COMPASS-SSH-TUNNEL');
function getConnectConfig(config) {
    const { localAddr, localPort, socks5Password, socks5Username, ...connectConfig } = config;
    return connectConfig;
}
function getSshTunnelConfig(config) {
    return {
        localAddr: '127.0.0.1',
        localPort: 0,
        socks5Username: undefined,
        socks5Password: undefined,
        ...config,
    };
}
let idCounter = 0;
class SshTunnel extends events_1.EventEmitter {
    constructor(config = {}) {
        super();
        this.connected = false;
        this.closed = false;
        this.connections = new Set();
        this.logCtx = `tunnel-${idCounter++}`;
        this.rawConfig = getSshTunnelConfig(config);
        this.sshClient = new ssh2_1.Client();
        this.sshClient.on('close', () => {
            log.info(mongoLogId(1001000252), this.logCtx, 'sshClient closed');
            this.connected = false;
        });
        this.forwardOut = (0, util_1.promisify)(this.sshClient.forwardOut.bind(this.sshClient));
        this.server = server_1.default.createServer(this.socks5Request.bind(this));
        if (this.rawConfig.socks5Username) {
            this.server.useAuth((0, UserPassword_1.default)((user, pass, cb) => {
                const success = this.rawConfig.socks5Username === user &&
                    this.rawConfig.socks5Password === pass;
                log.info(mongoLogId(1001000253), this.logCtx, 'Validated auth parameters', { success });
                queueMicrotask(() => cb(success));
            }));
        }
        else {
            log.info(mongoLogId(1001000254), this.logCtx, 'Skipping auth setup');
            this.server.useAuth((0, None_1.default)());
        }
        this.serverListen = (0, util_1.promisify)(this.server.listen.bind(this.server));
        this.serverClose = (0, util_1.promisify)(this.server.close.bind(this.server));
        for (const eventName of ['close', 'error', 'listening']) {
            this.server.on(eventName, this.emit.bind(this, eventName));
        }
    }
    get config() {
        const serverAddress = this.server.address();
        return {
            ...this.rawConfig,
            localPort: (typeof serverAddress !== 'string' && serverAddress?.port) ||
                this.rawConfig.localPort,
        };
    }
    async listen() {
        const { localPort, localAddr } = this.rawConfig;
        log.info(mongoLogId(1001000255), this.logCtx, 'Listening for Socks5 connections', { localAddr, localPort });
        await this.serverListen(localPort, localAddr);
        await this.connectSsh();
    }
    async close() {
        log.info(mongoLogId(1001000256), this.logCtx, 'Closing SSH tunnel');
        const [maybeError] = await Promise.all([
            this.serverClose().catch((e) => e),
            this.closeSshClient(),
            this.closeOpenConnections(),
        ]);
        if (maybeError) {
            throw maybeError;
        }
    }
    async connectSsh() {
        if (this.connected) {
            debug('already connected');
            return;
        }
        if (this.connectingPromise) {
            debug('reusing connectingPromise');
            return this.connectingPromise;
        }
        if (this.closed) {
            throw new Error('Disconnected.');
        }
        log.info(mongoLogId(1001000257), this.logCtx, 'Establishing new SSH connection');
        this.connectingPromise = Promise.race([
            (0, events_1.once)(this.sshClient, 'error').then(([err]) => {
                throw err;
            }),
            (() => {
                const waitForReady = (0, events_1.once)(this.sshClient, 'ready').then(() => { });
                this.sshClient.connect(getConnectConfig(this.rawConfig));
                return waitForReady;
            })(),
        ]);
        try {
            await this.connectingPromise;
        }
        catch (err) {
            this.emit('forwardingError', err);
            log.error(mongoLogId(1001000258), this.logCtx, 'Failed to establish new SSH connection', { error: err?.stack ?? String(err) });
            delete this.connectingPromise;
            await this.serverClose();
            throw err;
        }
        delete this.connectingPromise;
        this.connected = true;
        log.info(mongoLogId(1001000259), this.logCtx, 'Finished establishing new SSH connection');
    }
    async closeSshClient() {
        if (!this.connected) {
            return;
        }
        this.closed = true;
        const promise = (0, events_1.once)(this.sshClient, 'close');
        this.sshClient.end();
        return promise;
    }
    async closeOpenConnections() {
        const waitForClose = [];
        this.connections.forEach((socket) => {
            waitForClose.push((0, events_1.once)(socket, 'close'));
            socket.destroy();
        });
        await Promise.all(waitForClose);
        this.connections.clear();
    }
    async socks5Request(info, accept, deny) {
        const { srcAddr, srcPort, dstAddr, dstPort } = info;
        const logMetadata = { srcAddr, srcPort, dstAddr, dstPort };
        log.info(mongoLogId(1001000260), this.logCtx, 'Received Socks5 fowarding request', {
            ...logMetadata,
        });
        let socket = null;
        try {
            await this.connectSsh();
            let channel;
            try {
                channel = await this.forwardOut(srcAddr, srcPort, dstAddr, dstPort);
            }
            catch (err) {
                if (err.message === 'Not connected') {
                    this.connected = false;
                    log.error(mongoLogId(1001000261), this.logCtx, 'Error forwarding Socks5 request, retrying', {
                        ...logMetadata,
                        error: err.stack,
                    });
                    await this.connectSsh();
                    channel = await this.forwardOut(srcAddr, srcPort, dstAddr, dstPort);
                }
                else {
                    throw err;
                }
            }
            log.info(mongoLogId(1001000262), this.logCtx, 'Opened SSH channel and accepting socks5 request', {
                ...logMetadata,
            });
            socket = accept(true);
            this.connections.add(socket);
            socket.on('error', (err) => {
                log.error(mongoLogId(1001000263), this.logCtx, 'Error on Socks5 stream socket', {
                    ...logMetadata,
                    error: err.stack,
                });
                err.origin = err.origin ?? 'connection';
                this.emit('forwardingError', err);
            });
            socket.once('close', () => {
                log.info(mongoLogId(1001000264), this.logCtx, 'Socks5 stream socket closed', {
                    ...logMetadata,
                });
                this.connections.delete(socket);
            });
            socket.pipe(channel).pipe(socket);
        }
        catch (err) {
            this.emit('forwardingError', err);
            log.error(mongoLogId(1001000265), this.logCtx, 'Error establishing SSH channel for Socks5 request', {
                ...logMetadata,
                error: err.stack,
            });
            deny();
            if (socket) {
                err.origin = 'ssh-client';
                socket.destroy(err);
            }
        }
    }
}
exports.SshTunnel = SshTunnel;
exports.default = SshTunnel;
//# sourceMappingURL=index.js.map